# -*- coding: utf-8 -*-

"""
takumi_cli.deploy
~~~~~~~~~~~~~~~~~

Implement deploy command using ansible.
"""

import json
import os
import tempfile

PLAYBOOK_PATH = os.path.join(
    os.path.dirname(__file__), 'playbook', 'playbook.yml')


def _find_hosts(cwd):
    hosts = os.path.join(cwd, 'hosts')
    if os.path.exists(hosts):
        return hosts


def _vars(d):
    return ['{}={}'.format(k, v) for k, v in sorted(d.items())]


def _convert_section(name, data):
    hosts = data.get('hosts', {})
    group_vars = data.get('vars', {})

    # Set env variable
    group_vars['env'] = name

    items = ['[{}]'.format(name)]
    for host_name, host_vars in sorted(hosts.items()):
        item = [host_name]
        item.extend(_vars(host_vars))
        items.append(' '.join(item))

    section = '\n'.join(items)

    var_section = ''
    if group_vars:
        var_items = '\n'.join(_vars(group_vars))
        var_section = '[{}:vars]\n{}'.format(name, var_items)
    return [section, var_section]


def _convert_hosts(hosts):
    section_names = []
    sections = []
    for k, v in sorted(hosts.items()):
        converted = {}
        if not isinstance(v, dict):
            v = {'hosts': v}
        hs = v.get('hosts', [])
        if isinstance(hs, str):
            hs = [hs]
        for item in hs:
            if not isinstance(item, dict):
                item = {item: {}}
            converted.update(item)
        v['hosts'] = converted
        sections.extend(_convert_section(k, v))
        section_names.append(k)
    return sections, section_names


def _convert_crontab(data, app_name):
    working_dir = '/srv/{}'.format(app_name)
    items = []
    for item in data:
        sched = item.get('schedule')
        if isinstance(sched, str):
            minute, hour, day, month, weekday = sched.split()
            sched = {
                'minute': minute,
                'hour': hour,
                'day': day,
                'month': month,
                'weekday': weekday
            }
        item.update(sched)
        item['work_job'] = 'cd {} && {}'.format(working_dir, item['job'])
        items.append(item)
    return items


def _gen_hosts(deploy_config, app_name):
    hosts = deploy_config.get('targets', {})
    sections, section_names = _convert_hosts(hosts)

    crontab = _convert_crontab(deploy_config.get('crontab', []), app_name)
    deploy_vars = deploy_config.get('vars', {})
    deploy_vars.update({
        'app_name': app_name,
        'app_repo': os.getcwd(),
        'crontabs': json.dumps(crontab)
    })

    main_section = ['[service-deploy:children]']
    main_section.extend(section_names)
    sections.append('\n'.join(main_section))

    main_vars = ['[service-deploy:vars]']
    main_vars.extend(_vars(deploy_vars))
    sections.append('\n'.join(main_vars))

    hosts_file = tempfile.mktemp()
    with open(hosts_file, 'w') as f:
        f.write('\n'.join(sections))
    return hosts_file


def _compose_args(args):
    ansible_args = args['<ansible_args>']
    target = args['<target>']
    tags = args['--tags']

    if target:
        ansible_args.extend(['--limit', target])
    if tags:
        ansible_args.extend(['--tags', tags])

    def _has_inventory():
        return bool(set(('-i', '--inventory-file')).intersection(ansible_args))

    if target and not _has_inventory():
        from takumi_config import config
        hosts = _find_hosts(os.getcwd())
        if not hosts and config.deploy:
            hosts = _gen_hosts(config.deploy, config.app_name)
        if hosts:
            ansible_args.extend(['--inventory-file', hosts])

    if target:
        ansible_args.append(args['--play'] or PLAYBOOK_PATH)
    return ansible_args


def start(args):
    """Start deployment powered by ansible.

    :param args: arguments passed to ansible
    """
    args = _compose_args(args)
    args.insert(0, 'ansible-playbook')
    from ansible.cli.playbook import PlaybookCLI
    cli = PlaybookCLI(args)
    cli.parse()
    cli.run()



# -*- coding: utf-8 -*-

"""Show help for specific command.

Usage:
    takumi help <command>
    takumi help -h | --help

Options:
    -h, --help      Show this message and exit
"""

from ._base import parse_args


def run(args):
    args = parse_args('help', __doc__, args, {'<command>': str})
    return args['<command>']



# -*- coding: utf-8 -

"""
takumi_cli.runner
~~~~~~~~~~~~~~~~~

This module implements the `serve` command. This command is used for running
Takumi services using gunicorn gevent worker.

Available hooks:

    - after_load    Hook to be called after app imported
"""

import os
import socket
import sys

import gunicorn.workers
from gunicorn.app.base import Application
from gunicorn.config import Setting, validate_pos_int
from gunicorn.util import import_app
from gunicorn.workers.ggevent import GeventWorker

from takumi_config import config


class Worker(GeventWorker):
    def handle(self, listener, client, addr):
        from takumi.service import TakumiService
        from thriftpy.transport import TSocket

        thrift_service = TakumiService()
        ctx = thrift_service.context
        ctx.info['client_addr'] = addr[0]
        ctx.info['client_port'] = addr[1]
        ctx.info['worker'] = self

        client_timeout = self.app.cfg.client_timeout
        if client_timeout is not None:
            client.settimeout(client_timeout)

        handler_getter = self.app.wsgi()
        thrift_service.set_handler(handler_getter())
        sock = TSocket()
        sock.set_handle(client)

        logger = self.log
        try:
            thrift_service.run(sock)
        except socket.timeout:
            logger.warning('Client timeout: %r', addr)
        except socket.error as e:
            import errno
            if e.args[0] == errno.ECONNRESET:
                logger.debug('%r: %s', addr, str(e))
            elif e.args[0] == errno.EPIPE:
                logger.warning('%r: %s', addr, str(e))
            else:
                logger.exception('%r: %s', addr, str(e))
        except Exception as e:
            logger.exception('%r: %s', addr, str(e))

# Replace gunicorn default workers
gunicorn.workers.SUPPORTED_WORKERS.clear()
gunicorn.workers.SUPPORTED_WORKERS[
    'gevent_thriftpy'] = 'takumi_cli.runner.Worker'


class ClientTimeout(Setting):
    name = "client_timeout"
    cli = ["--client-timeout"]
    validator = validate_pos_int
    default = None
    desc = """\
        Seconds to timeout a client if client is silent after this duration
    """


class AppRunner(Application):
    def chdir(self):
        # chdir to the configured path before loading,
        # default is the current dir
        os.chdir(self.cfg.chdir)

        # add the path to sys.path
        sys.path.insert(0, self.cfg.chdir)

    def load(self):
        self.chdir()

        def load_app():
            app = import_app(config.app)
            app.hook_registry.on_after_load()
            return app
        return load_app

    def set_cfg(self):
        self.cfg.set('default_proc_name', config.app_name)
        self.cfg.set('worker_class', 'gevent_thriftpy')
        self.cfg.set('worker_connections', config.worker_connections)
        self.cfg.set('loglevel', 'info')
        self.cfg.set('graceful_timeout', 3)
        self.cfg.set('timeout', config.timeout)
        self.cfg.set('bind', '0.0.0.0:{}'.format(config.port))
        self.cfg.set('workers', config.workers)

        if config.env == 'dev' or config.syslog_disabled:
            self.cfg.set('errorlog', '-')
        else:
            self.cfg.set('syslog', True)
            self.cfg.set('syslog_facility', 'local6')
            self.cfg.set('syslog_addr', 'unix:///dev/log#dgram')

    def init(self, parser, opts, args):
        self.set_cfg()
        self.cfg.set('client_timeout', config.client_timeout)
        self.patch_gunicorn()

    def patch_gunicorn(self):
        import gunicorn.sock

        def _socket_str(self):
            return '%s:%d' % self.sock.getsockname()
        gunicorn.sock.TCPSocket.__str__ = _socket_str



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


setup(
    name='takumi_cli',
    version='0.1.1',
    description='Takumi service framework command line toolkit',
    long_description=open("README.rst").read(),
    author="Eleme Lab",
    author_email="imaralla@icloud.com",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    url='https://github.com/elemecreativelab/takumi-cli',
    install_requires=[
        'takumi',
        'takumi-ext',
        'takumi-config',
        'docopt',
        'schema',
        'gevent>=1.2.1',
        'thriftpy',
        'gunicorn',
    ],
    extras_require={
        'deploy': ['ansible>=2.2.0.0'],
        'shell': ['takumi-client', 'IPython']
    },
    entry_points={
        'console_scripts': [
            'takumi = takumi_cli.cmds:main'
        ]
    }
)
